#include <stdio.h>
#include <stdlib.h>
#include <timerstack.h>
#include <ui.h>
#include <jmp.h>
#include <jvmpi.h>

static JVMPI_RawMonitor timerstacks_monitor;
static int timerstacks_output_active = 0;
static int timerstacks_locked = 0;

timerstack* timerstack_new (size_t size) {
    char buf[64];
    timerstack *ts;

    if ((ts = (timerstack *)malloc(sizeof (*ts))) == NULL)
	return NULL;

    if ((ts->times = (methodtime *)malloc(sizeof (*ts->times) * size)) == NULL) {
	timerstack_free (ts);
	return NULL;
    }

    ts->top = 0;
    ts->max = size;
    ts->last_contentation = -1;
    snprintf(buf, 64, "_timerstack %p", ts);
    /* XXX Do we need to strdup this? */    
    ts->monitor = jmp_create_monitor(buf);
    return ts;
}


void timerstack_free (timerstack* s) {
    if (s == NULL)
	return;

    free (s->times);
    jmp_delete_monitor (s->monitor);
    free (s);
}


/** Expand the given stack.
 * It should be locked by the caller!
 */
void timerstack_expand (timerstack* s) {
    methodtime* ns;

    if (s == NULL)
	return;

    ns = realloc (s->times, sizeof (*s->times) * s->max * 2);
    if (ns) {
	s->times = ns;
	s->max *= 2;
    } else {
	fprintf (stderr, "timerstack_expand: failed to realloc methodtimes.\n");
    }
}

void timerstacks_lock_from_worker () {
    /* Mark timerstacks as locked */
    timerstacks_locked++;
    
    /* has timerstacks_lock_from_output been called? */
    if (timerstacks_output_active) {
	/* Oops, were already locked so undo the last change */
	timerstacks_locked--;

	/*
	 * We will enter the monitor after timerstacks_unlock_from_output
	 * has been called.
	 */
	jmp_lock_monitor (timerstacks_monitor);
	/* Lock again */
	timerstacks_locked++;
	jmp_unlock_monitor (timerstacks_monitor);
    }
}

void timerstacks_unlock_from_worker () {
    /* Mark timerstacks as unlocked */
    timerstacks_locked--;

    /*
     * If we are the last working thread that unlocks and
     * timerstacks_lock_from_output has been called wake up the waiting
     * output thread.
     */
    if (timerstacks_output_active && !timerstacks_locked) {
	jmp_monitor_notify_all (timerstacks_monitor);
    }
}

void timerstacks_lock_from_output () {
    jmp_lock_monitor (timerstacks_monitor);
    timerstacks_output_active = 1;
    
    /* Loop as long as worker threads have locked "individual" timerstacks. */
    while (timerstacks_locked) {
	jmp_monitor_wait (timerstacks_monitor, 0);
    }
}

void timerstacks_unlock_from_output () {
    timerstacks_output_active = 0;
    jmp_unlock_monitor (timerstacks_monitor);
}

void timerstack_init () {
    timerstacks_monitor = jmp_create_monitor ("_timerstacks");
}

/* Emacs Local Variables: */
/* Emacs mode:C */
/* Emacs c-indentation-style:"gnu" */
/* Emacs c-hanging-braces-alist:((brace-list-open)(brace-entry-open)(defun-open after)(substatement-open after)(block-close . c-snug-do-while)(extern-lang-open after)) */
/* Emacs c-cleanup-list:(brace-else-brace brace-elseif-brace space-before-funcall) */
/* Emacs c-basic-offset:4 */
/* Emacs End: */
